import math
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from codes.clu_stress import CluStress
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import confusion_matrix
from sklearn.datasets import make_blobs
from codes.earthquake_fms import set_rakes

variables = ['x', 'y', 'z'] #, 'r']
data = pd.read_csv('synthetic_fms_020421_labeled.csv')
data_true = data.copy()
data = data[variables]
data_x = data.x.copy()
data_y = data.y.copy()
if 'r' in variables:
    data = set_rakes(data)
data['PID'] = data_true.index.copy()
data_true['PID'] = data_true.index.copy()
data['cluster_fl'] = 0
scaler = MinMaxScaler()
data[variables] = scaler.fit_transform(data[variables])
if 'r' in variables:
    data.r = 1.5*data.r
clu_stress_obj = CluStress(data)
[data, data_outlier, clu_stress_obj.dist] = clu_stress_obj.select_outliers(data, k=20)
data = clu_stress_obj.commence_clustering(data, variables)
data = pd.concat([data, data_outlier])

data['x_raw'] = data_x
data['y_raw'] = data_y
# n_clusters_h = data.cluster_fl.max()
# if n_clusters_h == -1:
#     n_clusters_h = 'noise'
# ax = plt.subplot(1,1,1)
# scatter = plt.scatter(data['x_raw'],data['y_raw'], c=data.cluster_fl, cmap='rainbow')
# plt.xlabel('x')
# plt.ylabel('y')
# ax.set_xlim(xmin=0, xmax=300)
# plt.title('Automatized with ' + str(n_clusters_h) + ' clusters')
# plt.rcParams.update({'font.size': 15})
# legend = ax.legend(*scatter.legend_elements(),
#                    loc="upper right", title="Classes", prop={"size":10})
# ax.add_artist(legend)
# plt.savefig('CluStress_test.png')
# plt.show()

#
cluster_centres_x = []
cluster_centres_y = []
labels = []
for c_label in list(data_true.Cluster.unique()):
    cluster_centres_x.append(data_true[data_true.Cluster==c_label].x.mean())
    cluster_centres_y.append(data_true[data_true.Cluster==c_label].y.mean())
    labels.append(c_label)
cluster_centres_true = pd.DataFrame({'centre_x': cluster_centres_x, 'centre_y': cluster_centres_y, 'Cluster': labels})
print(cluster_centres_true)
#
#
cluster_centres_x = []
cluster_centres_y = []
labels = []
for c_label in list(data.cluster_fl.unique()):
    cluster_centres_x.append(data[data.cluster_fl==c_label].x_raw.mean())
    cluster_centres_y.append(data[data.cluster_fl==c_label].y_raw.mean())
    labels.append(c_label)
cluster_centres_pred = pd.DataFrame({'centre_x': cluster_centres_x, 'centre_y': cluster_centres_y, 'cluster_fl': labels})
print(cluster_centres_pred)
#
def dist_euclidean(x1, x2, y1, y2):
    return math.sqrt((x1-x2)**2 + (y1-y2)**2)

cluster_centres = cluster_centres_pred
cluster_centres['nearest_cluster_label'] = -2
cluster_centres['distance'] = 0
for cluster_flag in list(cluster_centres.cluster_fl.values):
    cluster_centres_part = cluster_centres[cluster_centres.cluster_fl==cluster_flag]
    distances = []
    flags = []
    for ind in list(cluster_centres_true.index):
        true_centre_x = cluster_centres_true.centre_x.iloc[ind]
        true_centre_y = cluster_centres_true.centre_y.iloc[ind]
        cluster = cluster_centres_true.Cluster.iloc[ind]
        distances.append(dist_euclidean(cluster_centres_part.centre_x, true_centre_x, cluster_centres_part.centre_y, true_centre_y))
        flags.append(cluster)
    dist_data = pd.DataFrame({'dist': distances, 'flag': flags})
    print(dist_data.flag.loc[dist_data.dist==dist_data.dist.min()].values[0])
    label_value = dist_data.flag.loc[dist_data.dist==dist_data.dist.min()].values[0]
    cluster_centres.nearest_cluster_label.loc[cluster_centres.cluster_fl==cluster_flag] = label_value
    cluster_centres.distance.loc[cluster_centres.cluster_fl==cluster_flag] = dist_data.dist.min()

for true_cluster_label in list(cluster_centres_true.Cluster.values):
    cluster_centres_partition = cluster_centres[cluster_centres.nearest_cluster_label==true_cluster_label]
    cond_1 = cluster_centres.nearest_cluster_label==true_cluster_label
    cond_2 = cluster_centres.distance!=cluster_centres_partition.distance.min()
    cluster_centres.nearest_cluster_label.loc[cond_1 & cond_2] = -2
cluster_centres.drop(columns=['distance'])
print(cluster_centres)
data['new_cluster_fl'] = -2
for cluster_flag in list(cluster_centres.cluster_fl.values):
    nearest_label = cluster_centres[cluster_centres.cluster_fl==cluster_flag].nearest_cluster_label.values[0]
    if nearest_label != -2:
        data.new_cluster_fl.loc[data.cluster_fl==cluster_flag] = nearest_label

data.new_cluster_fl.loc[data.new_cluster_fl==-2] = data.new_cluster_fl.max() + 1
print(data.new_cluster_fl.unique())

n_clusters_h = data.new_cluster_fl.max()
if n_clusters_h == -1:
    n_clusters_h = 'noise'
ax = plt.subplot(1,1,1)
scatter = plt.scatter(data['x_raw'],data['y_raw'], c=data.new_cluster_fl, cmap='rainbow')
plt.xlabel('x')
plt.ylabel('y')
ax.set_xlim(xmin=0, xmax=300)
plt.title('Automatized with ' + str(n_clusters_h) + ' clusters')
plt.rcParams.update({'font.size': 15})
legend = ax.legend(*scatter.legend_elements(),
                   loc="upper right", title="Classes", prop={"size":10})
ax.add_artist(legend)
plt.savefig('CluStress_test.png')
plt.show()

data = data.sort_values(by='PID')
data_true = data_true.sort_values(by='PID')
cm = confusion_matrix(data_true['Cluster'], data['new_cluster_fl'])
print('Confusion Matrix : \n', cm)

total=sum(sum(cm))
accuracy=(np.trace(cm))/total
print('Accuracy : ', accuracy)

categorical_accuracies = []
for ii in range(cm.shape[0]):
    categorical_accuracies.append(cm[ii,ii]/(sum(cm[ii,:])+sum(cm[:,ii])-cm[ii,ii]))

print(categorical_accuracies)




import itertools
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix


class DistanceMatricesNew:

    #TODO: this method works only with scaled coordinate values,
    # needs to be refactored to handle dimensional values as well
    # outlier ranges need to be computed for the full distance matrix
    def compute_distance_matrix(self, df, variables):
        DIST = np.zeros((len(df), len(df)))
        var = 0
        for variable in variables:
            coord_values_arr = df[variable].values
            coord_values_arr = np.tile(coord_values_arr, (len(coord_values_arr), 1))
            coord_values_arr_tr = coord_values_arr.T
            DIST += (coord_values_arr - coord_values_arr_tr)**2
            var = var + 1
        return DIST

    # converts distance_matrix to a sparse array dictionary of keys format
    def transform_distance_matrix(self, distance_matrix):
        dist_sparse = csr_matrix(np.triu(distance_matrix).copy())
        dist_dict = dist_sparse.todok()
        dist_dict = dict(dist_dict.items())
        dist_df = pd.DataFrame({'id_col': list(dist_dict.keys()), 'distance': list(dist_dict.values())})
        dist_df['id_0'] = dist_df['id_col'].apply(lambda x: x[0])
        dist_df['id_1'] = dist_df['id_col'].apply(lambda x: x[1])
        dist_df['id_0_r'] = dist_df['id_col'].apply(lambda x: x[1])
        dist_df['id_1_r'] = dist_df['id_col'].apply(lambda x: x[0])
        return dist_df

    # keeps only those elements of the distance matrix which are not clusterized
    def set_distance_matrix(self, dist, clusterized_list):
        all_clusterized_pairs = list(itertools.combinations(clusterized_list, 2))
        dist = dist.loc[np.invert(dist.id_col.isin(all_clusterized_pairs))]
        return dist

    def bin_distances(self, dist):
        dist = dist.sort_values(by='distance')
        var_bins = np.histogram_bin_edges(dist['distance'].values, bins='auto')
        #Interpolate between bins (resampling)
        for step in range(1):
            n = len(var_bins)
            var_bin_range = np.arange(0, 2*n, 2)
            var_bin_range_new = np.arange(2*n-1)
            var_bins = np.interp(var_bin_range_new, var_bin_range, var_bins)
        print(len(var_bins))
        dist['distance'] = list(np.digitize(dist['distance'].values, var_bins))
        dist = dist.sort_values(by='id_0')
        return dist

    def find_nearest_dists(self, dist):
        """Finds nearest distances for in the distance matrix to determine outlier ranges."""
        nearest_dists = []
        for row in range(dist.shape[0]):
            dist_i = dist[row,:]
            nearest_dists.append(np.min(dist_i[np.nonzero(dist_i)]))
        nearest_dists = pd.DataFrame({'distance': nearest_dists})
        return nearest_dists

    def flip_values(self, tuple_of_vals):
        flipped_list = [0,0]
        flipped_list[1] = tuple_of_vals[0]
        flipped_list[0] = tuple_of_vals[1]
        flipped_tuple = tuple(flipped_list)
        return flipped_tuple







#set the geologically meaningful rake categories based on the rake angle
def set_rakes(pontok):
    pontok.r.loc[pontok.r==0] = 1 #left_strike_slip
    pontok.r.loc[(pontok.r>0) & (pontok.r<20)] = 1 #left lateral strike-slip
    pontok.r.loc[(pontok.r>=20) & (pontok.r<70)] = 12 #reverse left-lateral oblique
    pontok.r.loc[(pontok.r>=70) & (pontok.r<110)] = 13 #reverse
    pontok.r.loc[(pontok.r>=110) & (pontok.r<160)] = 14 #reverse right-lateral oblique
    pontok.r.loc[(pontok.r>=160) & (pontok.r<=180)] = 1 #left lateral strike-slip
    pontok.r.loc[pontok.r==180] = 2 #right_strike_slip
    r_list_n = [0, -20, -70, -110, -160, -180]
    pontok.r.loc[((pontok.r>r_list_n[1]) & (pontok.r<=r_list_n[0]))] = 1 #left lateral strike-slip
    pontok.r.loc[((pontok.r>r_list_n[2]) & (pontok.r<=r_list_n[1]))] = 22 #normal left-lateral oblique
    pontok.r.loc[((pontok.r>r_list_n[3]) & (pontok.r<=r_list_n[2]))] = 23 #normal
    pontok.r.loc[((pontok.r>r_list_n[4]) & (pontok.r<=r_list_n[3]))] = 24 #normal right-lateral oblique
    pontok.r.loc[((pontok.r>=r_list_n[5]) & (pontok.r<=r_list_n[4]))] = 1 #left lateral strike-slip
    pontok.r.loc[((pontok.r>180-r_list_n[0]) & (pontok.r<=180-r_list_n[1]))] = 2 #right lateral strike-slip
    pontok.r.loc[((pontok.r>180-r_list_n[1]) & (pontok.r<=180-r_list_n[2]))] = 22 #normal left-lateral oblique
    pontok.r.loc[((pontok.r>180-r_list_n[2]) & (pontok.r<=180-r_list_n[3]))] = 23 #normal
    pontok.r.loc[((pontok.r>180-r_list_n[3]) & (pontok.r<=180-r_list_n[4]))] = 24 #normal right-lateral oblique
    pontok.r.loc[((pontok.r>=180-r_list_n[4]) & (pontok.r<=180-r_list_n[5]))] = 1 #left lateral strike-slip
    pontok.r.loc[(pontok.r==270) | (pontok.r==-90)] = 21 #normal dip-slip
    pontok.r.loc[pontok.r==90] = 11 #reverse-dip-slip
    return pontok


import pandas as pd
from unittest import TestCase
from src.clustress.clu_stress import CluStress


class TestCluStressNearestPoints(TestCase):
    mock_mindices0 = [0, 0, 0, 1, 1, 7, 7, 7, 5, 5, 5, 9]
    mock_mindices1 = [2, 5, 1, 4, 3, 9, 8, 7, 6, 5, 4, 10]
    mock_points = pd.DataFrame({'x': mock_mindices0, 'y': mock_mindices1})
    mock_points['PID'] = list(mock_points.index)
    mock_cluster_fl = 0
    mock_points['cluster_fl'] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    mock_variables = ['x', 'y']
    clu_stress = CluStress(mock_points)

    def test_group_connected_pairs_of_min_dist(self):
        check_cluster_fl = [1, 0, 1, 2, 2, 3, 3, 3, 4, 4, 4, 0]
        pontok, cluster_fl = self.clu_stress.clusterize_nearest(self.mock_points, self.mock_cluster_fl)
        self.assertListEqual(check_cluster_fl, list(pontok.cluster_fl))


class TestCluStressAllNearestPoints(TestCase):
    mock_mindices0 = [0, 0, 0, 1, 1, 7, 7, 7, 5, 5, 5, 9]
    mock_mindices1 = [2, 5, 1, 4, 3, 9, 8, 7, 6, 5, 4, 10]
    mock_points = pd.DataFrame({'x': mock_mindices0, 'y': mock_mindices1})
    mock_points['PID'] = list(mock_points.index)
    mock_cluster_fl = 0
    mock_points['cluster_fl'] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    mock_variables = ['x', 'y']
    clu_stress = CluStress(mock_points)

    def test_group_all_connected_pairs_of_min_dist(self):
        check_cluster_fl = dict({0: {'cluster_fl': 1}, 1: {'cluster_fl': 0}, 2: {'cluster_fl': 1}, 3: {'cluster_fl': 2},
                                 4: {'cluster_fl': 2}, 5: {'cluster_fl': 3}, 6: {'cluster_fl': 3}, 7: {'cluster_fl': 3},
                                 8: {'cluster_fl': 4}, 9: {'cluster_fl': 4}, 10: {'cluster_fl': 4}, 11: {'cluster_fl': 0}})
        pontok, cluster_fl = self.clu_stress.clusterize_nearest(self.mock_points.sample(frac=1), self.mock_cluster_fl)
        pontok.index = pontok.PID
        pontok = pontok.sort_index()
        self.assertEqual(check_cluster_fl, pontok[['cluster_fl']].to_dict(orient='index'))


class TestCluStressConnPoints(TestCase):
    mock_mindices0 = [0, 0, 0, 1, 1, 7, 7, 7, 5, 5, 5, 9]
    mock_mindices1 = [2, 5, 1, 4, 3, 9, 8, 7, 6, 5, 4, 10]
    mock_points = pd.DataFrame({'x': mock_mindices0, 'y': mock_mindices1})
    mock_points['PID'] = list(mock_points.index)
    mock_cluster_fl = 0
    mock_points['cluster_fl'] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    mock_variables = ['x', 'y']
    clu_stress = CluStress(mock_points)

    def test_get_connected_points(self):
        mock_nearest_point_pairs = [(0,2), (3,4), (5,6), (6,7), (8,9), (9,10)]
        connected_point_pairs = self.clu_stress.get_connected_points(mock_nearest_point_pairs)
        print(connected_point_pairs)


class TestCluStressCheckPoints(TestCase):
    mock_mindices0 = [0, 0, 0, 1, 1, 7, 7, 7, 5, 5, 5, 9]
    mock_mindices1 = [2, 5, 1, 4, 3, 9, 8, 7, 6, 5, 4, 10]
    mock_points = pd.DataFrame({'x': mock_mindices0, 'y': mock_mindices1})
    mock_points['PID'] = list(mock_points.index)
    mock_cluster_fl = 0
    mock_points['cluster_fl'] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    mock_variables = ['x', 'y']
    clu_stress = CluStress(mock_points)

    def test_check_if_nearest_point_is_part_of_a_cluster(self):
        mock_points_tmp = self.mock_points.copy()
        mock_points_tmp['cluster_fl'] = [1, 0, 1, 2, 2, 3, 3, 3, 4, 4, 4, 0]
        mock_connected_points = pd.DataFrame({'id_col': [(1,3)], 'id_0': [1], 'id_1': [3], 'distance_0': [2.0], 'distance': [2.0], 'distance_1': [0.0]})
        mock_connected_points['cluster_fl'] = 0
        mock_clusterized = mock_points_tmp.loc[mock_points_tmp['cluster_fl'] != 0]
        to_merge = self.clu_stress.check_if_nearest_point_is_part_of_a_cluster(mock_clusterized, mock_connected_points, self.mock_variables)
        print(to_merge)


class TestCluStressLinkPoints(TestCase):
    mock_mindices0 = [0, 0, 0, 1, 1, 7, 7, 7, 5, 5, 5, 9]
    mock_mindices1 = [2, 5, 1, 4, 3, 9, 8, 7, 6, 5, 4, 10]
    mock_points = pd.DataFrame({'x': mock_mindices0, 'y': mock_mindices1})
    mock_points['PID'] = list(mock_points.index)
    mock_cluster_fl = 0
    mock_points['cluster_fl'] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    mock_variables = ['x', 'y']
    clu_stress = CluStress(mock_points)

    def test_link_points_which_have_their_nearest_neighbours_in_a_cluster(self):
        self.clu_stress.link_points_which_have_their_nearest_neighbours_in_a_cluster(self.mock_points, self.mock_variables)


class TestCluStressMergePoints(TestCase):
    mock_mindices0 = [0, 0, 0, 1, 1, 7, 7, 7, 5, 5, 5, 9]
    mock_mindices1 = [2, 5, 1, 4, 3, 9, 8, 7, 6, 5, 4, 10]
    mock_points = pd.DataFrame({'x': mock_mindices0, 'y': mock_mindices1})
    mock_points['PID'] = list(mock_points.index)
    mock_cluster_fl = 0
    mock_points['cluster_fl'] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    mock_variables = ['x', 'y']
    clu_stress = CluStress(mock_points)

    def test_merge_points(self):
        mock_to_merge = pd.DataFrame({'id_col': [(1,3)], 'id_0': [1], 'id_1': [3], 'distance_0': [2.0], 'distance': [2.0], 'distance_1': [0.0]})
        mock_to_merge['cluster_fl'] = 2
        mock_to_merge['cluster_fl_1'] = mock_to_merge['cluster_fl']
        mock_to_merge = mock_to_merge.drop(columns=['cluster_fl', 'id_col', 'distance_0', 'distance', 'distance_1'],
                                 axis=1)
        mock_to_merge = self.clu_stress.select_largest_nearest_cluster(mock_to_merge, self.mock_points)
        points = self.clu_stress.merge_points(self.mock_points, mock_to_merge)
        print(points)


class TestCluStressFullClustering(TestCase):
    mock_mindices0 = [0, 0, 0, 1, 1, 7, 7, 7, 5, 5, 5, 9]
    mock_mindices1 = [2, 5, 1, 4, 3, 9, 8, 7, 6, 5, 4, 10]
    mock_points = pd.DataFrame({'x': mock_mindices0, 'y': mock_mindices1})
    mock_points['PID'] = list(mock_points.index)
    mock_cluster_fl = 0
    mock_points['cluster_fl'] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    mock_variables = ['x', 'y']
    clu_stress = CluStress(mock_points)

    def test_commence_clustering(self):
        points = self.clu_stress.commence_clustering(self.mock_points, self.mock_variables)
        print(points)
        # import matplotlib.pyplot as plt
        # plt.scatter(points['x'], points['y'], c=points.cluster_fl)
        # plt.show()


class TestCluStressFullClustering_b(TestCase):
    mock_mindices0 = [0, 1, 2, 3, 5, 8, 8, 8, 8, 14]
    mock_mindices1 = [3, 1, 3, 0, 3, 3, 2, 4, 6, 8]
    mock_points = pd.DataFrame({'x': mock_mindices0, 'y': mock_mindices1})
    mock_points['PID'] = list(mock_points.index)
    mock_cluster_fl = 0
    mock_points['cluster_fl'] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    mock_variables = ['x', 'y']
    clu_stress = CluStress(mock_points)

    def test_commence_clustering_b(self):
        [points, points_outlier, _] = self.clu_stress.select_outliers(self.mock_points)
        points = self.clu_stress.commence_clustering(self.mock_points, self.mock_variables)
        print(points)
        import matplotlib.pyplot as plt
        plt.scatter(points['x'], points['y'], c=points.cluster_fl)
        plt.show()

class TestBlobs(TestCase):
    from sklearn.datasets import make_blobs, make_moons
    var, clust = make_blobs(n_samples=100, centers=3, n_features=2)
    #var, clust = make_moons(n_samples=100, noise=0.1)
    mock_variables = ['x', 'y']

    def test_blobs(self):
        import numpy as np
        points = pd.DataFrame({'x': list(self.var[:, 0]), 'y': list(self.var[:, 1]), 'cluster_fl': np.zeros((1,100)).tolist()[0], 'cluster_ref': list(self.clust)})
        from sklearn.preprocessing import MinMaxScaler
        scaler = MinMaxScaler()
        for column in self.mock_variables:
            points[column] = scaler.fit_transform(points[column].values.reshape(-1, 1))
        points['PID'] = points.index
        print(points.head())
        clu_stress = CluStress(points[['x','y', 'cluster_fl', 'PID']])
        [points, points_outlier, _] = clu_stress.select_outliers(points)
        points = clu_stress.commence_clustering(points[['x', 'y', 'cluster_fl', 'PID']], self.mock_variables)
        print(points)
        import matplotlib.pyplot as plt
        points = pd.concat([points_outlier, points], axis=0)
        n_clusters_h = len(points.cluster_fl.unique())
        fig, ax = plt.subplots()
        scatter = ax.scatter(points['x'], points['y'], c=points.cluster_fl)
        legend = ax.legend(*scatter.legend_elements(num=n_clusters_h), loc='lower right', title='Blob clusters',
                           prop={'size': 4})
        ax.add_artist(legend)
        plt.show()


class TestOutliers(TestCase):
    mock_mindices0 = [0, 0, 0, 1, 1, 7, 7, 7, 5, 5, 5, 9, 30, 15]
    mock_mindices1 = [2, 5, 1, 4, 3, 9, 8, 7, 6, 5, 4, 10, 50, 25]
    ids = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,11, 12, 13]
    mock_points = pd.DataFrame({'x': mock_mindices0, 'y': mock_mindices1})
    from sklearn.preprocessing import MinMaxScaler
    scaler = MinMaxScaler()
    for column in mock_points.columns:
        mock_points[column] = scaler.fit_transform(mock_points[column].values.reshape(-1, 1))
    mock_points['PID'] = list(mock_points.index)
    mock_points['cluster_fl'] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    mock_variables = ['x', 'y']
    clu_stress = CluStress(mock_points)

    def test_outliers(self):
        [points, points_outlier, _] = self.clu_stress.select_outliers(self.mock_points)
        points = pd.concat([points_outlier, points], axis=0)
        import matplotlib.pyplot as plt
        plt.scatter(points['x'], points['y'], c=points.cluster_fl)
        plt.show()


import pandas as pd
import matplotlib.pyplot as plt
from src.clustress.clu_stress import CluStress
from sklearn.preprocessing import MinMaxScaler
from sklearn.datasets import make_blobs

## Synthetic blobs
# -------------------------------------------------------------------------------------------- #

pontok = make_blobs(n_samples=100, n_features=2, centers=3, random_state=0, cluster_std=0.4)
scaler = MinMaxScaler()
pontok_x = scaler.fit_transform(pontok[0][:,0].reshape(-1, 1)).flatten()
pontok_y = scaler.fit_transform(pontok[0][:,1].reshape(-1, 1)).flatten()
data = pd.DataFrame({'x': list(pontok_x), 'y': list(pontok_y)})
data['cluster_fl'] = 0
data['PID'] = data.index
#data = pd.read_csv('./test_data.csv')
#print(data.head())
variables = data.columns[:-2]
cols = data.columns
data = data[cols]
print(variables)
print(cols)

clu_stress_obj = CluStress(data)
print(data)
[data, data_outlier, clu_stress_obj.dist] = clu_stress_obj.select_outliers(data)
print(data_outlier)
pontok = clu_stress_obj.commence_clustering(data, variables)

n_clusters_h = len(pontok.cluster_fl.unique())
if n_clusters_h == -1:
    n_clusters_h = 'noise'
ax = plt.subplot(1, 1, 1)
plt.scatter(pontok['x'], pontok['y'], c=pontok.cluster_fl, cmap='rainbow')
ax.set_aspect(1)
plt.title('CluStress automated with ' + str(n_clusters_h) + ' clusters')
plt.rcParams.update({'font.size': 15})
plt.savefig('CluStress_Blobs.png')
plt.show()

## Vrancea earthquake FMS
# -------------------------------------------------------------------------------------------- #





import numpy as np

class DistanceMatrices:
    def DistanceMatrix2D(self, df):
        DIST = np.zeros((len(df), len(df)))
        for ii in range(len(df)):
            for jj in range(len(df)):
                d_x_ij = df['x'].iloc[ii] - df['x'].iloc[jj];
                d_y_ij = df['y'].iloc[ii] - df['y'].iloc[jj];
                DIST[ii, jj] = d_x_ij ** 2 + d_y_ij ** 2;

        return DIST

    def DistanceMatrix3D(self, df):
        DIST = np.zeros((len(df), len(df)))
        for ii in range(len(df)):
            for jj in range(len(df)):
                d_x_ij = df['x'].iloc[ii] - df['x'].iloc[jj];
                d_y_ij = df['y'].iloc[ii] - df['y'].iloc[jj];
                d_z_ij = df['z'].iloc[ii] - df['z'].iloc[jj];
                DIST[ii, jj] = d_x_ij ** 2 + d_y_ij ** 2 + d_z_ij ** 2;

        return DIST

    def DistanceMatrix4D(self, df):
        DIST = np.zeros((len(df), len(df)))
        for ii in range(len(df)):
            for jj in range(len(df)):
                d_x_ij = df['x'].iloc[ii] - df['x'].iloc[jj];
                d_y_ij = df['y'].iloc[ii] - df['y'].iloc[jj];
                d_z_ij = df['z'].iloc[ii] - df['z'].iloc[jj];
                d_r_ij = df['r'].iloc[ii] - df['r'].iloc[jj];
                DIST[ii, jj] = d_x_ij ** 2 + d_y_ij ** 2 + d_z_ij ** 2 + d_r_ij ** 2;

        return DIST

    def DiscretizeDistances(self, CluStressObj, DIST):
        # Diszkretizalt bin-ek keszitesehez sorbarendezzuk a tavolsag-ertekeket
        DIST_list = list(DIST.flatten())
        DIST_list.sort()

        # Kozuluk megtartjuk csak az egyedi ertekeket
        set_of_unique_dist_values = set(DIST_list)

        dist_value_list = list(set_of_unique_dist_values)
        dist_value_list.sort()

        # Az adathalmaz diffuzivitasat a tavolsag-ertekek szorasakent definialjuk
        overall_diffusivity = np.std(np.array(dist_value_list))

        collection_of_ids_to_unify = {}
        collection_key = 0
        id_tuple = ()

        # Ciklus: a tavolsag-ertkek egysegbe rendezesehez
        for distance in dist_value_list:
            if distance != 0:
                id_tuple = np.where(((DIST - distance) ** 2 < 100 * overall_diffusivity ** 2) & (DIST != 0))
            if (len(id_tuple) != 0) and (len(id_tuple[0]) != 0):
                collection_of_ids_to_unify.update({collection_key: [id_tuple, distance]})
                collection_key = collection_key + 1
        for collection_key in list(collection_of_ids_to_unify.keys()):
            id_tuple = collection_of_ids_to_unify[collection_key][0]
            distance = collection_of_ids_to_unify[collection_key][1]
            DIST[id_tuple] = distance

        return DIST


import setuptools

requires = ['pandas>=1.1.5',
            'scikit-learn>=0.0',
            'scipy>=1.7.3',
            'matplotlib>=3.5.1',
            'xlrd>=2.0.1']

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="clu-stress",
    version="1.2.0",
    author="Lukacs Kuslits",
    author_email="kuslits.lukacs@epss.hun-ren.hu",
    description="A non-parametric clustering algorithm.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lukacskuslits/clu-stress",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5",
    install_requires=requires,
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
)

# package_dir={"": "clustress"},
# packages=setuptools.find_packages(where="clustress"),


import math
import sys
import time
import numpy as np
import pandas as pd
import scipy.sparse as sp
from scipy.sparse import csr_matrix
from scipy.stats import iqr
from sklearn.preprocessing import MinMaxScaler
from sklearn.datasets import make_blobs
from sklearn.metrics import silhouette_score
from sklearn.metrics import silhouette_samples
from clustress.distance_matrices_new import DistanceMatricesNew
import itertools
from itertools import compress
import matplotlib.pyplot as plt


class CluStress:

	def __init__(self, points):
		col_list = list(points.columns)
		print(col_list)
		if 'PID' not in col_list:
			print('Column PID missing from input data!')
			sys.exit(-1)
		if 'cluster_fl' not in col_list:
			print('Column cluster_fl missing from input data!')
			sys.exit(-1)
		dist_array = DistanceMatricesNew().compute_distance_matrix(points, col_list[:-2])
		self.dist_array = dist_array
		dist = DistanceMatricesNew().transform_distance_matrix(dist_array)
		dist['rev_id_col'] = dist['id_col'].apply(DistanceMatricesNew().flip_values)
		self.dist = dist

	def bin_distances(self):
		self.dist = self.dist.sort_values(by='distance')
		var_bins = np.histogram_bin_edges(self.dist['distance'].values, bins='auto')
		#Interpolate between bins (resampling)
		for step in range(1):
			n = len(var_bins)
			var_bin_range = np.arange(0, 2*n, 2)
			var_bin_range_new = np.arange(2*n-1)
			var_bins = np.interp(var_bin_range_new, var_bin_range, var_bins)
		print(len(var_bins))
		self.dist['distance'] = list(np.digitize(self.dist['distance'].values, var_bins))
		self.dist = self.dist.sort_values(by='id_0')

	def get_outlier_distances(self, k):
		"""Tukey's fence outlier detection."""
		nearest_dists = DistanceMatricesNew().find_nearest_dists(self.dist_array)
		quant_df = nearest_dists.distance.quantile([.25,.75])
		q_1 = quant_df.iloc[0]
		q_3 = quant_df.iloc[1]
		outlier_range = q_3 + k*(q_3 - q_1)
		return outlier_range

	def select_outliers(self, points, k=1.5):
		outlier_range = self.get_outlier_distances(k)
		print('Outlier distance:')
		print(outlier_range)
		dist_to_sort = self.dist.copy()
		for _id in list(points['PID']):
			dist_temp = dist_to_sort.loc[(dist_to_sort['id_0'] == _id) | (dist_to_sort['id_1'] == _id)].sort_values(by=['distance'])
			nearest_dist_of_point = dist_temp['distance'].iloc[0]
			if nearest_dist_of_point > outlier_range:
				points.cluster_fl.loc[points.PID == _id] = -1
		self.dist['cluster_fl'] = self.dist.id_0.map(points.set_index('PID')['cluster_fl'])
		self.dist['cluster_fl_1'] = self.dist.id_1.map(points.set_index('PID')['cluster_fl'])
		self.dist.cluster_fl.loc[self.dist.id_1 == points.PID.max()] = list(self.dist.cluster_fl_1.loc[self.dist.id_1 == points.PID.max()].astype(int))
		self.dist.drop(columns=['cluster_fl_1'])
		dist_filtered = self.dist.loc[self.dist['cluster_fl']!=-1].copy()
		dist_filtered = dist_filtered.drop(columns=['cluster_fl', 'cluster_fl_1'])
		self.dist = dist_filtered
		points_filtered = points.loc[points['cluster_fl'] != -1].copy()
		points_outlier = points.loc[points['cluster_fl'] == -1].copy()
		return points_filtered, points_outlier, dist_filtered

	def transform_distance_matrix(self, distance_matrix):
		"""Converts distance_matrix to a sparse array dictionary of keys format."""
		dist_sparse = csr_matrix(np.triu(distance_matrix).copy())
		dist_dict = dist_sparse.todok()
		dist_dict = dict(dist_dict.items())
		dist_df = pd.DataFrame({'id_col': list(dist_dict.keys()), 'distance': list(dist_dict.values())})
		dist_df['id_0'] = dist_df['id_col'].apply(lambda x: x[0])
		dist_df['id_1'] = dist_df['id_col'].apply(lambda x: x[1])
		return dist_df

	def set_distance_matrix(self, clusterized_list):
		"""Keeps only those elements of the distance matrix which are not clusterized."""
		all_clusterized_pairs = list(itertools.combinations(clusterized_list, 2))
		self.dist = self.dist.loc[np.invert(self.dist.id_col.isin(all_clusterized_pairs))]
		self.dist = self.dist.loc[np.invert(self.dist.rev_id_col.isin(all_clusterized_pairs))]

	def find_connected_pairs(self, point_set, point_pairs):
		connected_pairs = ()
		for point in point_set:
			is_connected_pair = [point in i for i in point_pairs]
			connected_pairs += tuple(compress(point_pairs, is_connected_pair))
			point_pairs = list(compress(point_pairs, np.invert(is_connected_pair)))
			if len(point_pairs)==0:
				break
		return connected_pairs, point_pairs

	def connect_connected_pairs(self, point_set, connected_pairs):
		for connected_pair in connected_pairs:
			point_set += connected_pair
			point_set = tuple(set(point_set))
		return point_set

	def connect_nearest_points_to_pair(self, point_set, point_pairs):
		while True:
			point_set_old = point_set
			[connected_pairs, point_pairs] = self.find_connected_pairs(point_set, point_pairs)
			point_set = self.connect_connected_pairs(point_set, connected_pairs)
			if (len(point_set_old) == len(point_set)) or (len(point_pairs)==0):
				break
		return point_set, point_pairs

	def get_connected_points(self, nearest_point_pairs):
		remaining_point_pairs = nearest_point_pairs
		connected_point_pairs = []
		increment = 0
		L = len(nearest_point_pairs)
		while True:
			point_pair = nearest_point_pairs[increment]
			[connected_point_set, remaining_point_pairs] = self.connect_nearest_points_to_pair(point_pair, remaining_point_pairs)
			nearest_point_pairs = [connected_point_set] + remaining_point_pairs
			connected_point_pairs.append(connected_point_set)
			L=len(nearest_point_pairs)
			increment = increment+1
			if increment >= L:
				if len(remaining_point_pairs)!=0:
					[connected_point_set, _] = self.connect_nearest_points_to_pair(remaining_point_pairs[0], remaining_point_pairs)
					connected_point_pairs.append(connected_point_set)
				break
		return connected_point_pairs

	def clusterize_nearest(self, points, cluster_fl):
		if cluster_fl != 0:
			clusterized = points.loc[(points['cluster_fl'] != 0)].reset_index(drop=True)
		else:
			clusterized = points.loc[points['cluster_fl'] != cluster_fl].reset_index(drop=True)
		self.set_distance_matrix(list(clusterized.PID))
		clusterized_id = list(clusterized.PID)
		disttemp = self.dist.loc[~(self.dist.id_0.isin(clusterized_id)) & ~(self.dist.id_1.isin(clusterized_id))].copy()
		indexes_of_nearest_points = list(disttemp.loc[(disttemp.distance == disttemp.distance.min())].id_col)
		connected_nearest_points = self.get_connected_points(indexes_of_nearest_points)
		for connected_point_set in connected_nearest_points:
			cluster_fl = int(cluster_fl + 1)
			points['cluster_fl'][points.PID.isin(list(connected_point_set))] = int(cluster_fl)
			clusterized = points.loc[points['cluster_fl'] == cluster_fl].reset_index(drop=True)
			self.set_distance_matrix(list(clusterized.PID))
		points = points.reset_index(drop=True)
		return points, cluster_fl

	def check_if_nearest_point_is_part_of_a_cluster(self, clusterized, to_merge, variables):
		to_merge = self.join_with_clusterized(clusterized, to_merge, 'id_1', variables)
		to_merge['point_id'] = to_merge.index
		to_merge = to_merge[to_merge.cluster_fl > 0]
		return to_merge

	def join_with_clusterized(self, clusterized, to_merge, point_id_to_check, variables):
		to_merge = to_merge.set_index(point_id_to_check)
		to_merge = to_merge.drop(columns=['cluster_fl'], axis=1).join(
		clusterized.drop(columns=variables, axis=1).set_index('PID'), how='left')
		to_merge = to_merge.fillna(0)
		return to_merge

	def merge_points(self, points, to_merge):
		points['point_id'] = points['PID']
		if 0 in list(points.columns):
			points = points.drop(columns=[0], axis=1)
		if 0 in list(to_merge.columns):
			to_merge = to_merge.drop(columns=[0], axis=1)
		points = points.set_index('point_id').join(to_merge.set_index('id_0'), how='left')
		points = points.fillna(0)
		points['cluster_fl'] = points['cluster_fl'] + points['cluster_fl_1']
		points['cluster_fl'] = points['cluster_fl'].astype(int)
		points = points.drop(columns=['cluster_fl_1'], axis=1)
		return points

	def link_points_which_have_their_nearest_neighbours_in_a_cluster(self, points, variables):
		not_clusterized = points.loc[points['cluster_fl'] == 0]
		not_clusterized_id = list(not_clusterized.PID)
		point_id = list(points.PID.astype(int))
		clusterized = points.loc[points['cluster_fl'] != 0]
		not_clust_pairs = [(nc, p) for nc, p in itertools.product(not_clusterized_id, point_id)]
		id_0 = list(itertools.chain.from_iterable(itertools.repeat(x, len(point_id)) for x in not_clusterized_id))
		id_1 = point_id * len(not_clusterized_id)
		nearest_neighbour_table = pd.DataFrame({'id_col': not_clust_pairs, 'id_0': id_0, 'id_1': id_1})
		nearest_neighbour_table['distance'] = nearest_neighbour_table.id_col.map(self.dist.set_index('id_col')['distance'])
		nearest_neighbour_table['distance_0'] = nearest_neighbour_table['distance']
		nearest_neighbour_table = nearest_neighbour_table.drop(columns=['distance'], axis=1)
		nearest_neighbour_table['distance'] = nearest_neighbour_table.id_col.map(self.dist.set_index('rev_id_col')['distance'])
		nearest_neighbour_table['distance_1'] = nearest_neighbour_table['distance']
		nearest_neighbour_table = nearest_neighbour_table.fillna(0)
		nearest_neighbour_table['distance'] = nearest_neighbour_table['distance'] + nearest_neighbour_table['distance_0']
		nearest_neighbour_table = nearest_neighbour_table.loc[nearest_neighbour_table.distance > 0]
		to_merge = nearest_neighbour_table[nearest_neighbour_table.groupby('id_0')\
										  ['distance'].transform(min) == nearest_neighbour_table['distance']]
		to_merge['cluster_fl'] = int(0)
		to_merge = self.check_if_nearest_point_is_part_of_a_cluster(clusterized, to_merge, variables)
		to_merge['cluster_fl_1'] = to_merge['cluster_fl'].astype(int)
		to_merge = to_merge.drop(columns=['cluster_fl', 'point_id', 'id_col', 'distance_0', 'distance', 'distance_1'], axis=1)
		to_merge = self.select_largest_nearest_cluster(to_merge, points)
		points = self.merge_points(points, to_merge)
		return points

	def select_largest_nearest_cluster(self, to_merge, points):
		points_with_more_than_one_nearest_neighbour = set(to_merge['id_0'].loc[to_merge['id_0'].duplicated()==True])
		if len(points_with_more_than_one_nearest_neighbour) > 0:
			for point_w_more_neighbours in points_with_more_than_one_nearest_neighbour:
				to_merge_tmp = to_merge.loc[to_merge.id_0 == point_w_more_neighbours]
				to_merge_tmp['halmaz_meret'] = 0
				to_merge_tmp = to_merge_tmp.reset_index(drop=True)
				for point_to_merge_more_neigh in range(len(to_merge_tmp)):
					to_merge_tmp.iloc[point_to_merge_more_neigh].halmaz_meret = len(points.loc[points.cluster_fl==to_merge_tmp.iloc[point_to_merge_more_neigh].cluster_fl_1])
				to_merge_tmp = to_merge_tmp.sort_values(by=['halmaz_meret'], ascending=False).iloc[0]
				to_merge = to_merge.loc[to_merge.id_0 != point_w_more_neighbours]
				to_merge = pd.concat([to_merge, to_merge_tmp])
				if 'halmaz_meret' in list(to_merge.columns):
					to_merge = to_merge.drop(columns=['halmaz_meret'], axis=1)
		return to_merge

	def commence_clustering(self, points, variables):
		cluster_fl = int(0)
		set_of_flags = {0}
		zeros_now = [0]
		zeros_before = [0,0]
		points = points.reset_index(drop=True)
		self.dist = DistanceMatricesNew().bin_distances(self.dist)
		while (len(zeros_now) < len(zeros_before)) and (0 in set_of_flags):
			self.dist.to_csv('dist_py_1.csv')
			zeros_before = list(points.loc[points.cluster_fl==0].cluster_fl)
			print('zeros before:')
			print(zeros_before)
			[points, cluster_fl] = self.clusterize_nearest(points, cluster_fl)
			self.dist.to_csv('dist_py_2.csv')
			points = self.link_points_which_have_their_nearest_neighbours_in_a_cluster(points, variables)
			self.dist.to_csv('dist_py_3.csv')
			zeros_now = list(points.loc[points.cluster_fl==0].cluster_fl)
			print('zeros now:')
			print(zeros_now)
			set_of_flags = set(points.cluster_fl)
		return points
